@@ -1,18 +1,14 @@
 #
 # This file is part of pyasn1 software.
 #
-# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
-# License: http://snmplabs.com/pyasn1/license.html
+# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
+# License: https://pyasn1.readthedocs.io/en/latest/license.html
 #
 import math
 import pickle
+import platform
 import sys
-
-try:
-    import unittest2 as unittest
-
-except ImportError:
-    import unittest
+import unittest
 
 from tests.base import BaseTestCase
 
@@ -146,23 +142,17 @@ class NoValueTestCase(BaseTestCase):
         else:
             assert False, 'boolean evaluation works for NoValue object'
     
+    @unittest.skipIf(
+        platform.python_implementation() == "PyPy",
+        "getsizeof() raises TypeError on PyPy"
+    )
     def testSizeOf(self):
         try:
-            if hasattr(sys, 'getsizeof'):
-                sys.getsizeof(univ.noValue)
-
-            # TODO: remove when Py2.5 support is gone
-            elif sys.version_info > (2, 6):
-                raise unittest.SkipTest("no sys.getsizeof() method")
+            sys.getsizeof(univ.noValue)
 
         except PyAsn1Error:
             assert False, 'sizeof failed for NoValue object'
 
-        except TypeError:
-            # TODO: remove when Py2.5 support is gone
-            if sys.version_info > (2, 6):
-                raise unittest.SkipTest("sys.getsizeof() raises TypeError")
-
 
 class IntegerTestCase(BaseTestCase):
     def testStr(self):
@@ -278,9 +268,8 @@ class IntegerTestCase(BaseTestCase):
     def testCeil(self):
         assert math.ceil(univ.Integer(1)) == 1, '__ceil__() fails'
 
-    if sys.version_info[0:2] > (2, 5):
-        def testTrunc(self):
-            assert math.trunc(univ.Integer(1)) == 1, '__trunc__() fails'
+    def testTrunc(self):
+        assert math.trunc(univ.Integer(1)) == 1, '__trunc__() fails'
 
     def testPrettyIn(self):
         assert univ.Integer('3') == 3, 'prettyIn() fails'
@@ -437,9 +426,8 @@ class BitStringTestCase(BaseTestCase):
         assert self.b.clone("'A98A'H")[1] == 0
         assert self.b.clone("'A98A'H")[2] == 1
 
-    if sys.version_info[:2] > (2, 4):
-        def testReverse(self):
-            assert list(reversed(univ.BitString([0, 0, 1]))) == list(univ.BitString([1, 0, 0]))
+    def testReverse(self):
+        assert list(reversed(univ.BitString([0, 0, 1]))) == list(univ.BitString([1, 0, 0]))
 
     def testAsOctets(self):
         assert self.b.clone(hexValue='A98A').asOctets() == ints2octs((0xa9, 0x8a)), 'testAsOctets() fails'
@@ -539,9 +527,8 @@ class OctetStringWithUnicodeMixIn(object):
         assert self.encodedPythonString in s
         assert self.encodedPythonString * 2 not in s
 
-    if sys.version_info[:2] > (2, 4):
-       def testReverse(self):
-           assert list(reversed(univ.OctetString(self.encodedPythonString))) == list(reversed(self.encodedPythonString))
+    def testReverse(self):
+        assert list(reversed(univ.OctetString(self.encodedPythonString))) == list(reversed(self.encodedPythonString))
 
 
 class OctetStringWithAsciiTestCase(OctetStringWithUnicodeMixIn, BaseTestCase):
@@ -565,11 +552,6 @@ class OctetStringUnicodeErrorTestCase(BaseTestCase):
         except PyAsn1UnicodeEncodeError:
             pass
 
-        # TODO: remove when Py2.5 support is gone
-        else:
-            if sys.version_info > (2, 6):
-                assert False, 'Unicode encoding error not caught'
-
     def testDecodeError(self):
         serialized = ints2octs((0xff, 0xfe))
 
@@ -585,11 +567,6 @@ class OctetStringUnicodeErrorTestCase(BaseTestCase):
         except PyAsn1UnicodeDecodeError:
             pass
 
-        # TODO: remove when Py2.5 support is gone
-        else:
-            if sys.version_info > (2, 6):
-                assert False, 'Unicode decoding error not caught'
-
 
 class OctetStringWithUtf8TestCase(OctetStringWithUnicodeMixIn, BaseTestCase):
     initializer = (208, 176, 208, 177, 208, 178)
@@ -860,9 +837,8 @@ class RealTestCase(BaseTestCase):
     def testCeil(self):
         assert math.ceil(univ.Real(1.2)) == 2.0, '__ceil__() fails'
 
-    if sys.version_info[0:2] > (2, 5):
-        def testTrunc(self):
-            assert math.trunc(univ.Real(1.1)) == 1.0, '__trunc__() fails'
+    def testTrunc(self):
+        assert math.trunc(univ.Real(1.1)) == 1.0, '__trunc__() fails'
 
     def testTag(self):
         assert univ.Real().tagSet == tag.TagSet(
